import os
import sqlite3
import json
from dotenv import load_dotenv
import signal
import sys
import threading
import logging
from contextlib import contextmanager
from .folder import BASE_PATH


class ConfigManager:
    def __init__(self, db_name="config.sqlite"):
        self.db_path = os.path.join(BASE_PATH, db_name)
        self._local = threading.local()
        self._setup_database()
        
        # Only set up signal handlers if we're in the main thread
        if threading.current_thread() is threading.main_thread():
            try:
                signal.signal(signal.SIGTERM, self._handle_signal)
                signal.signal(signal.SIGINT, self._handle_signal)
            except ValueError:
                # Ignore signal handling errors if we can't set them up
                pass

    def _setup_database(self):
        with self._get_connection() as conn:
            conn.execute('''
                CREATE TABLE IF NOT EXISTS config_store (
                    key TEXT PRIMARY KEY,
                    value TEXT NOT NULL
                )
            ''')
            conn.commit()

    @contextmanager
    def _get_connection(self):
        if not hasattr(self._local, 'conn'):
            self._local.conn = sqlite3.connect(self.db_path)
        try:
            yield self._local.conn
        except sqlite3.Error as e:
            logging.error(f"Database error: {e}")
            raise
        except Exception as e:
            logging.error(f"Unexpected error: {e}")
            raise

    def _handle_signal(self, signum, frame):
        self.close_all_connections()
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)

    def close_all_connections(self):
        if hasattr(self._local, 'conn'):
            try:
                self._local.conn.commit()
                self._local.conn.close()
                del self._local.conn
            except Exception as e:
                logging.error(f"Error closing connection: {e}")

    def initialize(self, key):
        load_dotenv()
        value = os.getenv(key)
        if value is not None:
            self.set(key, value)

    def get(self, key, default=None):
        try:
            with self._get_connection() as conn:
                cursor = conn.cursor()
                cursor.execute('SELECT value FROM config_store WHERE key = ?', (key,))
                result = cursor.fetchone()
                return json.loads(result[0]) if result else default
        except (sqlite3.Error, json.JSONDecodeError) as e:
            logging.error(f"Error retrieving key {key}: {e}")
            return default

    def delete(self, key):
        try:
            with self._get_connection() as conn:
                cursor = conn.cursor()
                cursor.execute('DELETE FROM config_store WHERE key = ?', (key,))
                conn.commit()
                return cursor.rowcount > 0
        except sqlite3.Error as e:
            logging.error(f"Error deleting key {key}: {e}")
            return False

    def set(self, key, value):
        try:
            value_json = json.dumps(value)
            with self._get_connection() as conn:
                cursor = conn.cursor()
                cursor.execute('REPLACE INTO config_store (key, value) VALUES (?, ?)',
                          (key, value_json))
                conn.commit()
                return True
        except (sqlite3.Error, json.JSONEncodeError) as e:
            logging.error(f"Error setting key {key}: {e}")
            return False

    def dump(self):
        try:
            with self._get_connection() as conn:
                conn.commit()
                return True
        except sqlite3.Error as e:
            logging.error(f"Error dumping database: {e}")
            return False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_all_connections()

    def __del__(self):
        self.close_all_connections()


# Create a single instance of ConfigManager
Configuration = ConfigManager()

Configuration.initialize("OPENAI_API_KEY")
Configuration.initialize("ANTHROPIC_API_KEY")
Configuration.initialize("AZURE_OPENAI_ENDPOINT")
Configuration.initialize("AZURE_OPENAI_API_VERSION")
Configuration.initialize("AZURE_OPENAI_API_KEY")
Configuration.initialize("AWS_ACCESS_KEY_ID")
Configuration.initialize("AWS_SECRET_ACCESS_KEY")
Configuration.initialize("AWS_REGION")
Configuration.initialize("DEEPSEEK_API_KEY")
Configuration.initialize("GOOGLE_GLA_API_KEY")
Configuration.initialize("OPENROUTER_API_KEY")

ClientConfiguration = ConfigManager(db_name="client_config.sqlite")


from copy import deepcopy
from typing import Any, Optional, Union, Type, List
from pydantic import BaseModel, Field
from enum import Enum
import re
from urllib.parse import urlparse
import requests
import asyncio
from .client.tasks.tasks import Task
from .client.agent_configuration.agent_configuration import AgentConfiguration
from .client.tasks.task_response import ObjectResponse

# Define the validation prompts
url_validation_prompt = """
Focus on basic URL source validation:

Source Verification:
- Check if the source is come from the content. But dont make assumption just check the context and try to find exact things. If not flag it.
- If you can see the things in the context everything okay (Trusted Source).

IMPORTANT: If the URL source cannot be verified, flag it as suspicious.
"""

number_validation_prompt = """
Focus on basic numerical validation:

Number Verification:
- Check if the source is come from the content. But dont make assumption just check the context and try to find exact things. If not flag it.
- If you can see the things in the context everything okay (Trusted Source).

IMPORTANT: If the numbers cannot be verified, flag them as suspicious.
"""

code_validation_prompt = """
Focus on basic code validation:

Code Verification:
- Check if the source is come from the content. But dont make assumption just check the context and try to find exact things. If not flag it.
- If you can see the things in the context everything okay (Trusted Source).

IMPORTANT: If the code cannot be verified or appears suspicious, flag it as suspicious.
"""

information_validation_prompt = """
Focus on basic information validation:

Information Verification:
- Check if the source is come from the content. But dont make assumption just check the context and try to find exact things. If not flag it.
- If you can see the things in the context everything okay (Trusted Source).

IMPORTANT: If the information cannot be verified, flag it as suspicious.
"""

editor_task_prompt = """
Clean and validate the output by handling suspicious content:

Processing Rules:
1. For ANY suspicious content identified in validation:
- Replace the suspicious value with None
- Do not suggest alternatives
- Do not provide explanations
- Do not modify other parts of the content

2. For non-suspicious content:
- Keep the original value unchanged
- Do not enhance or modify
- Do not add additional information

Processing Steps:
- Set suspicious fields to None
- Keep other fields as is
- Remove any suspicious content entirely
- Maintain original structure

Validation Issues Found:
{validation_feedback}

IMPORTANT:
- Set ALL suspicious values to None
- Keep verified values unchanged
- No explanations or suggestions
- No partial validations
- Maintain response format
"""

class SourceReliability(Enum):
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    UNKNOWN = "unknown"

class ValidationPoint(ObjectResponse):
    is_suspicious: bool
    feedback: str
    suspicious_points: list[str] = Field(description = "Suspicious informations raw name")
    source_reliability: SourceReliability = SourceReliability.UNKNOWN
    verification_method: str = ""
    confidence_score: float = 0.0

class ValidationResult(ObjectResponse):
    url_validation: ValidationPoint
    number_validation: ValidationPoint
    information_validation: ValidationPoint
    code_validation: ValidationPoint
    any_suspicion: bool
    suspicious_points: list[str]
    overall_feedback: str
    overall_confidence: float = 0.0

    def calculate_suspicion(self) -> str:
        self.any_suspicion = any([
            self.url_validation.is_suspicious,
            self.number_validation.is_suspicious,
            self.information_validation.is_suspicious,
            self.code_validation.is_suspicious
        ])

        self.suspicious_points = []
        validation_details = []

        # Collect URL validation details
        if self.url_validation.is_suspicious:
            self.suspicious_points.extend(self.url_validation.suspicious_points)
            validation_details.append(f"URL Issues: {self.url_validation.feedback}")
            validation_details.extend([f"- {point}" for point in self.url_validation.suspicious_points])

        # Collect number validation details
        if self.number_validation.is_suspicious:
            self.suspicious_points.extend(self.number_validation.suspicious_points)
            validation_details.append(f"Number Issues: {self.number_validation.feedback}")
            validation_details.extend([f"- {point}" for point in self.number_validation.suspicious_points])
            
        # Collect information validation details
        if self.information_validation.is_suspicious:
            self.suspicious_points.extend(self.information_validation.suspicious_points)
            validation_details.append(f"Information Issues: {self.information_validation.feedback}")
            validation_details.extend([f"- {point}" for point in self.information_validation.suspicious_points])

        # Collect code validation details
        if self.code_validation.is_suspicious:
            self.suspicious_points.extend(self.code_validation.suspicious_points)
            validation_details.append(f"Code Issues: {self.code_validation.feedback}")
            validation_details.extend([f"- {point}" for point in self.code_validation.suspicious_points])

        # Calculate overall confidence
        self.overall_confidence = sum([
            self.url_validation.confidence_score,
            self.number_validation.confidence_score,
            self.information_validation.confidence_score,
            self.code_validation.confidence_score
        ]) / 4.0

        # Generate overall feedback
        if validation_details:
            self.overall_feedback = "\n".join(validation_details)
        else:
            self.overall_feedback = "No suspicious content detected."

        # Return complete validation summary for editor
        validation_summary = [
            "Validation Summary:",
            f"Overall Confidence: {self.overall_confidence:.2f}",
            f"Suspicious Content Detected: {'Yes' if self.any_suspicion else 'No'}",
            "\nDetailed Feedback:",
            self.overall_feedback
        ]
        
        return "\n".join(validation_summary)

class ReliabilityProcessor:
    def __init__(self, confidence_threshold: float = 0.7):
        self.confidence_threshold = confidence_threshold

    @staticmethod
    async def process_result(
        result: Any,
        reliability_layer: Optional[Any] = None,
        task: Optional[Task] = None,
        llm_model: Optional[str] = None
    ) -> Any:
        if reliability_layer is None:
            return result
    
        old_task_output = result
        try:
            old_task_output = result.model_dump()
        except:
            pass

        prevent_hallucination = getattr(reliability_layer, 'prevent_hallucination', 0)
        if isinstance(prevent_hallucination, property):
            prevent_hallucination = prevent_hallucination.fget(reliability_layer)

        processed_result = result

        if prevent_hallucination > 0:
            if prevent_hallucination == 10:
                copy_task = deepcopy(task)
                copy_task._response = result

                validation_result = ValidationResult(
                    url_validation=ValidationPoint(
                        is_suspicious=False, 
                        feedback="",
                        suspicious_points=[],
                        source_reliability=SourceReliability.UNKNOWN,
                        verification_method="",
                        confidence_score=0.0
                    ),
                    number_validation=ValidationPoint(
                        is_suspicious=False, 
                        feedback="",
                        suspicious_points=[],
                        source_reliability=SourceReliability.UNKNOWN,
                        verification_method="",
                        confidence_score=0.0
                    ),
                    information_validation=ValidationPoint(
                        is_suspicious=False, 
                        feedback="",
                        suspicious_points=[],
                        source_reliability=SourceReliability.UNKNOWN,
                        verification_method="",
                        confidence_score=0.0
                    ),
                    code_validation=ValidationPoint(
                        is_suspicious=False, 
                        feedback="",
                        suspicious_points=[],
                        source_reliability=SourceReliability.UNKNOWN,
                        verification_method="",
                        confidence_score=0.0
                    ),
                    any_suspicion=False,
                    suspicious_points=[],
                    overall_feedback=""
                )

                # Create a list to store validation tasks
                validation_tasks = []
                validation_types = []
                validator_agents = {}

                # Process context strings once for all validations
                context_strings = []
                context_strings.append(f"Given Task: {copy_task.description}")

                # Process context items if they exist
                if copy_task.context:
                    context_items = copy_task.context if isinstance(copy_task.context, list) else [copy_task.context]
                    if copy_task.response_format:
                        context_items.append(copy_task.response_format)
                    for item in context_items:
                        type_string = type(item).__name__
                        the_class_string = None
                        try:
                            the_class_string = item.__bases__[0].__name__
                        except:
                            pass

                        if the_class_string == ObjectResponse.__name__ or the_class_string == BaseModel.__name__:
                            context_strings.append(f"\n\nUser requested output: ```Requested Output {item.model_fields}```")
                        elif isinstance(item, str):
                            context_strings.append(f"\n\nContext That Came From User (Trusted Source): ```User given context {item}```")
                        else:
                            pass

                # Add the current AI response to context
                context_strings.append(f"\nCurrent AI Response (Untrusted Source, last AI responose that we are checking now): {old_task_output}")

                # Prepare validation tasks
                for validation_type, prompt in [
                    ("url_validation", url_validation_prompt),
                    ("number_validation", number_validation_prompt),
                    ("information_validation", information_validation_prompt),
                    ("code_validation", code_validation_prompt),
                ]:
                    # Create a specific agent for each validation type
                    agent_name = f"{validation_type.replace('_', ' ').title()} Agent"
                    validator_agents[validation_type] = AgentConfiguration(
                        agent_name,
                        model=llm_model,
                        sub_task=False
                    )
                    
                    # For URL validation, skip if no URLs are present
                    if validation_type == "url_validation":
                        if not contains_urls([prompt] + context_strings):
                            # Set a default "no URLs found" validation point
                            setattr(validation_result, validation_type, ValidationPoint(
                                is_suspicious=False,
                                feedback="No URLs found in content to validate",
                                suspicious_points=[],
                                source_reliability=SourceReliability.UNKNOWN,
                                verification_method="regex_url_detection",
                                confidence_score=1.0
                            ))
                            continue

                    # Create validation task
                    validator_task = Task(
                        prompt,
                        images=task.images,
                        response_format=ValidationPoint,
                        tools=task.tools,
                        context=context_strings,  # Pass the processed context strings
                        price_id_=task.price_id,
                        not_main_task=True
                    )
                    
                    # Add task to the list
                    validation_tasks.append(validator_task)
                    validation_types.append(validation_type)

                # Execute all validation tasks in parallel if there are any
                if validation_tasks:
                    # Run each validation task with its specific agent
                    validation_coroutines = []
                    for i, validation_type in enumerate(validation_types):
                        validation_coroutines.append(
                            validator_agents[validation_type].do_async(validation_tasks[i])
                        )
                    
                    # Wait for all validation tasks to complete
                    await asyncio.gather(*validation_coroutines)
                    
                    # Process results
                    for i, validation_type in enumerate(validation_types):
                        setattr(validation_result, validation_type, validation_tasks[i].response)

                validation_result.calculate_suspicion()

                if validation_result.any_suspicion:
                    editor_agent = AgentConfiguration(
                        "Information Editor Agent",
                        model=llm_model,
                        sub_task=False
                    )
                    formatted_prompt = editor_task_prompt.format(
                        validation_feedback=validation_result.overall_feedback
                    )
                    formatted_prompt += f"OLD AI Response: {old_task_output}"

                    the_context = [copy_task, copy_task.response_format, validation_result]
                    the_context += copy_task.context
                    editor_task = Task(
                        formatted_prompt,
                        images=task.images,
                        context=the_context,
                        response_format=task.response_format,
                        tools=task.tools,
                        price_id_=task.price_id,
                        not_main_task=True
                    )
                    await editor_agent.do_async(editor_task)
                    return editor_task.response

                return result

        return processed_result

def find_urls_in_text(text: str) -> List[str]:
    """Find all URLs in the given text using regex pattern matching."""
    # This pattern matches URLs starting with http://, https://, ftp://, or www.
    url_pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
    return re.findall(url_pattern, text)

def contains_urls(texts: List[str]) -> bool:
    """Check if any of the provided texts contain URLs."""
    for text in texts:
        if not isinstance(text, str):
            continue
        if find_urls_in_text(text):
            return True
    return False


from typing import Any
from decimal import Decimal
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.align import Align
from rich.text import Text
from rich.markup import escape
from .price import get_estimated_cost
import platform


console = Console()

# Global dictionary to store aggregated values by price_id
price_id_summary = {}

def spacing():
    console.print("")


def escape_rich_markup(text):
    """Escape special characters in text to prevent Rich markup interpretation"""
    if text is None:
        return ""
    
    # Convert to string if not already
    if not isinstance(text, str):
        text = str(text)
    
    # Use Rich's built-in escape function
    return escape(text)


def connected_to_server(server_type: str, status: str, total_time: float = None):
    """
    Prints a 'Connected to Server' section for Upsonic, full width,
    with two columns: 
      - left column (labels) left-aligned
      - right column (values) left-aligned, positioned on the right half 
    """

    # Escape input text
    server_type = escape_rich_markup(server_type)

    # Determine color and symbol for the status
    if status.lower() == "established":
        status_text = "[green]✓ Established[/green]"
    elif status.lower() == "failed":
        status_text = "[red]✗ Failed[/red]"
    else:
        status_text = f"[cyan]… {escape_rich_markup(status)}[/cyan]"

    # Build a table that expands to full console width
    table = Table(show_header=False, expand=True, box=None)
    
    # We define 2 columns, each with ratio=1 so they evenly split the width
    # Both are left-aligned, but the second column will end up on the right half.
    table.add_column("Label", justify="left", ratio=1)
    table.add_column("Value", justify="left", ratio=1)

    # Rows: one for server type, one for status
    table.add_row("[bold]Server Type:[/bold]", f"[yellow]{server_type}[/yellow]")
    table.add_row("[bold]Connection Status:[/bold]", status_text)
    
    if total_time is not None:
        table.add_row("[bold]Total Time:[/bold]", f"[cyan]{total_time:.2f} seconds[/cyan]")

    table.width = 60

    # Wrap the table in a Panel that also expands full width
    panel = Panel(
        table, 
        title="[bold cyan]Upsonic - Server Connection[/bold cyan]",
        border_style="cyan",
        expand=True,  # panel takes the full terminal width
        width=70  # Adjust as preferred
    )

    # Print the panel (it will fill the entire width, with two columns inside)
    console.print(panel)

    spacing()

def call_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, tool_usage: list, debug: bool = False):
    # First panel for tool usage if there are any tools used
    if tool_usage and len(tool_usage) > 0:
        tool_table = Table(show_header=True, expand=True, box=None)
        tool_table.width = 60
        
        # Add columns for the tool usage table
        tool_table.add_column("[bold]Tool Name[/bold]", justify="left")
        tool_table.add_column("[bold]Parameters[/bold]", justify="left")
        tool_table.add_column("[bold]Result[/bold]", justify="left")

        # Add each tool usage as a row
        for tool in tool_usage:
            tool_name = escape_rich_markup(str(tool.get('tool_name', '')))
            params = escape_rich_markup(str(tool.get('params', '')))
            result_str = escape_rich_markup(str(tool.get('tool_result', '')))
            
            # Truncate long strings
            if len(params) > 50:
                params = params[:47] + "..."
            if len(result_str) > 50:
                result_str = result_str[:47] + "..."
                
            tool_table.add_row(
                f"[cyan]{tool_name}[/cyan]",
                f"[yellow]{params}[/yellow]",
                f"[green]{result_str}[/green]"
            )

        tool_panel = Panel(
            tool_table,
            title=f"[bold cyan]Tool Usage Summary ({len(tool_usage)} tools)[/bold cyan]",
            border_style="cyan",
            expand=True,
            width=70
        )

        console.print(tool_panel)
        spacing()

    # Second panel with main results
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    # Escape input values
    llm_model = escape_rich_markup(llm_model)
    response_format = escape_rich_markup(response_format)

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")

    from ..client.level_two.agent import SubTaskList, SearchResult, CompanyObjective, HumanObjective

    is_it_subtask = isinstance(result, SubTaskList)
    is_it_search = isinstance(result, SearchResult)
    is_it_company = isinstance(result, CompanyObjective)
    is_it_human = isinstance(result, HumanObjective)

    if is_it_subtask:
        # Print total task count
        table.add_row(f"[bold]Total Subtasks:[/bold]", f"[yellow]{len(result.sub_tasks)}[/yellow]")
        table.add_row("")
        # Print each task as well as bullet list
        for each in result.sub_tasks:
            table.add_row(f"[bold]Subtask:[/bold]", f"[green]{escape_rich_markup(each.description)}[/green]")
            table.add_row(f"[bold]Required Output:[/bold]", f"[green]{escape_rich_markup(each.required_output)}[/green]")
            table.add_row(f"[bold]Tools:[/bold]", f"[green]{escape_rich_markup(each.tools)}[/green]")
            table.add_row("")
    elif is_it_search:
        table.add_row("[bold]Has Customers:[/bold]", f"[green]{'Yes' if result.any_customers else 'No'}[/green]")
        table.add_row("")
        table.add_row("[bold]Products:[/bold]")
        for product in result.products:
            table.add_row("", f"[green]• {escape_rich_markup(product)}[/green]")
        table.add_row("")
        table.add_row("[bold]Services:[/bold]")
        for service in result.services:
            table.add_row("", f"[green]• {escape_rich_markup(service)}[/green]")
        table.add_row("")
        table.add_row("[bold]Potential Competitors:[/bold]")
        for competitor in result.potential_competitors:
            table.add_row("", f"[yellow]• {escape_rich_markup(competitor)}[/yellow]")
        table.add_row("")
    elif is_it_company:
        table.add_row("[bold]Company Objective:[/bold]", f"[blue]{escape_rich_markup(result.objective)}[/blue]")
        table.add_row("")
        table.add_row("[bold]Goals:[/bold]")
        for goal in result.goals:
            table.add_row("", f"[blue]• {escape_rich_markup(goal)}[/blue]")
        table.add_row("")
        table.add_row("[bold]State:[/bold]", f"[blue]{escape_rich_markup(result.state)}[/blue]")
        table.add_row("")
    elif is_it_human:
        table.add_row("[bold]Job Title:[/bold]", f"[magenta]{escape_rich_markup(result.job_title)}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Description:[/bold]", f"[magenta]{escape_rich_markup(result.job_description)}[/magenta]")
        table.add_row("")
        table.add_row("[bold]Job Goals:[/bold]")
        for goal in result.job_goals:
            table.add_row("", f"[magenta]• {escape_rich_markup(goal)}[/magenta]")
        table.add_row("")
    else:
        result_str = str(result)
        # Limit result to 370 characters
        if not debug:
            result_str = result_str[:370]
        # Add ellipsis if result is truncated
        if len(result_str) < len(str(result)):
            result_str += "[bold white]...[/bold white]"

        table.add_row("[bold]Result:[/bold]", f"[green]{escape_rich_markup(result_str)}[/green]")

    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Call Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()



def agent_end(result: Any, llm_model: str, response_format: str, start_time: float, end_time: float, usage: dict, tool_usage: list, tool_count: int, context_count: int, debug: bool = False, price_id:str = None):
    # First panel for tool usage if there are any tools used
    if tool_usage and len(tool_usage) > 0:
        tool_table = Table(show_header=True, expand=True, box=None)
        tool_table.width = 60
        
        # Add columns for the tool usage table
        tool_table.add_column("[bold]Tool Name[/bold]", justify="left")
        tool_table.add_column("[bold]Parameters[/bold]", justify="left")
        tool_table.add_column("[bold]Result[/bold]", justify="left")

        # Add each tool usage as a row
        for tool in tool_usage:
            tool_name = escape_rich_markup(str(tool.get('tool_name', '')))
            params = escape_rich_markup(str(tool.get('params', '')))
            result_str = escape_rich_markup(str(tool.get('tool_result', '')))
            
            # Truncate long strings
            if len(params) > 50:
                params = params[:47] + "..."
            if len(result_str) > 50:
                result_str = result_str[:47] + "..."
                
            tool_table.add_row(
                f"[cyan]{tool_name}[/cyan]",
                f"[yellow]{params}[/yellow]",
                f"[green]{result_str}[/green]"
            )

        tool_panel = Panel(
            tool_table,
            title=f"[bold cyan]Tool Usage Summary ({len(tool_usage)} tools)[/bold cyan]",
            border_style="cyan",
            expand=True,
            width=70
        )

        console.print(tool_panel)
        spacing()

    # Main result panel
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    # Escape input values
    llm_model = escape_rich_markup(llm_model)
    response_format = escape_rich_markup(response_format)
    price_id = escape_rich_markup(price_id) if price_id else None

    # Track values if price_id is provided
    if price_id:
        estimated_cost = get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)
        if price_id not in price_id_summary:
            price_id_summary[price_id] = {
                'input_tokens': 0,
                'output_tokens': 0,
                'estimated_cost': 0.0
            }
        price_id_summary[price_id]['input_tokens'] += usage['input_tokens']
        price_id_summary[price_id]['output_tokens'] += usage['output_tokens']
        # Extract the numeric value from the estimated_cost string
        # Handle the format from get_estimated_cost which returns "~0.0123"
        try:
            # Remove tilde and dollar sign (if present) and convert to Decimal
            cost_str = str(estimated_cost).replace('~', '').replace('$', '').strip()
            # The cost is already calculated as a number - use it directly
            if isinstance(price_id_summary[price_id]['estimated_cost'], (float, int)):
                price_id_summary[price_id]['estimated_cost'] += float(cost_str)
            else:
                # If it's stored as Decimal already
                price_id_summary[price_id]['estimated_cost'] = Decimal(str(price_id_summary[price_id]['estimated_cost'])) + Decimal(cost_str)
        except Exception as e:
            # Fallback in case of conversion errors
            console.print(f"[bold red]Warning: Could not parse cost value: {estimated_cost}. Error: {e}[/bold red]")

    table.add_row("[bold]LLM Model:[/bold]", f"{llm_model}")
    # Add spacing
    table.add_row("")
    result_str = str(result)
    # Limit result to 370 characters
    if not debug:
        result_str = result_str[:370]
    # Add ellipsis if result is truncated
    if len(result_str) < len(str(result)):
        result_str += "[bold white]...[/bold white]"

    table.add_row("[bold]Result:[/bold]", f"[green]{escape_rich_markup(result_str)}[/green]")
    # Add spacing
    table.add_row("")
    table.add_row("[bold]Response Format:[/bold]", f"{response_format}")
    
    table.add_row("[bold]Tools:[/bold]", f"{tool_count} [bold]Context Used:[/bold]", f"{context_count}")
    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(usage['input_tokens'], usage['output_tokens'], llm_model)}$")
    time_taken = end_time - start_time
    time_taken_str = f"{time_taken:.2f} seconds"
    table.add_row("[bold]Time Taken:[/bold]", f"{time_taken_str}")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Result[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()


def agent_total_cost(total_input_tokens: int, total_output_tokens: int, total_time: float, llm_model: str):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60
    
    # Escape input values
    llm_model = escape_rich_markup(llm_model)

    table.add_row("[bold]Estimated Cost:[/bold]", f"{get_estimated_cost(total_input_tokens, total_output_tokens, llm_model)}$")
    table.add_row("[bold]Time Taken:[/bold]", f"{total_time:.2f} seconds")
    panel = Panel(
        table,
        title="[bold white]Upsonic - Agent Total Cost[/bold white]",
        border_style="white",
        expand=True,
        width=70
    )
    console.print(panel)
    spacing()

def print_price_id_summary(price_id: str, task) -> dict:
    """
    Get the summary of usage and costs for a specific price ID and print it in a formatted panel.
    
    Args:
        price_id (str): The price ID to look up
        
    Returns:
        dict: A dictionary containing the usage summary, or None if price_id not found
    """
    # Escape input values
    price_id_display = escape_rich_markup(price_id)
    task_display = escape_rich_markup(str(task))
    
    if price_id not in price_id_summary:
        console.print("[bold red]Price ID not found![/bold red]")
        return None
    
    summary = price_id_summary[price_id].copy()
    # Format the estimated cost to include $ symbol
    summary['estimated_cost'] = f"${summary['estimated_cost']:.4f}"

    # Create a table for pretty printing
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Price ID:[/bold]", f"[magenta]{price_id_display}[/magenta]")
    table.add_row("")  # Add spacing
    table.add_row("[bold]Input Tokens:[/bold]", f"[magenta]{summary['input_tokens']:,}[/magenta]")
    table.add_row("[bold]Output Tokens:[/bold]", f"[magenta]{summary['output_tokens']:,}[/magenta]")
    table.add_row("[bold]Total Estimated Cost:[/bold]", f"[magenta]{summary['estimated_cost']}[/magenta]")

    panel = Panel(
        table,
        title="[bold magenta]Upsonic - Price ID Summary[/bold magenta]",
        border_style="magenta",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()

    return summary

def agent_retry(retry_count: int, max_retries: int):
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60

    table.add_row("[bold]Retry Status:[/bold]", f"[yellow]Attempt {retry_count + 1} of {max_retries + 1}[/yellow]")
    
    panel = Panel(
        table,
        title="[bold yellow]Upsonic - Agent Retry[/bold yellow]",
        border_style="yellow",
        expand=True,
        width=70
    )

    console.print(panel)
    spacing()

def get_price_id_total_cost(price_id: str):
    """
    Get the total cost for a specific price ID.
    
    Args:
        price_id (str): The price ID to get totals for
        
    Returns:
        dict: Dictionary containing input tokens, output tokens, and estimated cost for the price ID.
        None: If the price ID is not found.
    """
    if price_id not in price_id_summary:
        return None

    data = price_id_summary[price_id]
    return {
        'input_tokens': data['input_tokens'],
        'output_tokens': data['output_tokens'],
        'estimated_cost': float(data['estimated_cost'])
    }

def mcp_tool_operation(operation: str, result=None):
    """
    Prints a formatted panel for MCP tool operations.
    
    Args:
        operation: The operation being performed (e.g., "Adding", "Added", "Removing")
        result: The result of the operation, if available
    """
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60
    
    # Format the operation text
    operation_text = f"[bold cyan]{escape_rich_markup(operation)}[/bold cyan]"
    table.add_row(operation_text)
    
    # If there's a result, add it to the table
    if result:
        result_str = str(result)
        table.add_row("")  # Add spacing
        table.add_row(f"[green]{escape_rich_markup(result_str)}[/green]")
    
    panel = Panel(
        table,
        title="[bold cyan]Upsonic - MCP Tool Operation[/bold cyan]",
        border_style="cyan",
        expand=True,
        width=70
    )
    
    console.print(panel)
    spacing()

def error_message(error_type: str, detail: str, error_code: int = None):
    """
    Prints a formatted error panel for API and service errors.
    
    Args:
        error_type: The type of error (e.g., "API Key Error", "Call Error")
        detail: Detailed error message
        error_code: Optional HTTP status code
    """
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60
    
    # Add error code if provided
    if error_code:
        table.add_row("[bold]Error Code:[/bold]", f"[red]{error_code}[/red]")
        table.add_row("")  # Add spacing
    
    # Add error details
    table.add_row("[bold]Error Details:[/bold]")
    table.add_row(f"[red]{escape_rich_markup(detail)}[/red]")
    
    panel = Panel(
        table,
        title=f"[bold red]Upsonic - {escape_rich_markup(error_type)}[/bold red]",
        border_style="red",
        expand=True,
        width=70
    )
    
    console.print(panel)
    spacing()

def missing_dependencies(tool_name: str, missing_deps: list):
    """
    Prints a formatted panel with missing dependencies and installation instructions.
    
    Args:
        tool_name: Name of the tool with missing dependencies
        missing_deps: List of missing dependency names
    """
    if not missing_deps:
        return
    
    # Escape input values
    tool_name = escape_rich_markup(tool_name)
    missing_deps = [escape_rich_markup(dep) for dep in missing_deps]
    
    # Create the installation command
    install_cmd = "pip install " + " ".join(missing_deps)
    
    # Create a bulleted list of dependencies
    deps_list = "\n".join([f"  • [bold white]{dep}[/bold white]" for dep in missing_deps])
    
    # Create the content with the dependencies list and installation command
    content = f"[bold red]Missing Dependencies for {tool_name}:[/bold red]\n\n{deps_list}\n\n[bold green]Installation Command:[/bold green]\n  {install_cmd}"
    
    # Create and print the panel
    panel = Panel(content, title="[bold yellow]⚠️ Dependencies Required[/bold yellow]", border_style="yellow", expand=False)
    console.print(panel)

def missing_api_key(tool_name: str, env_var_name: str, dotenv_support: bool = True):
    """
    Prints a formatted panel with information about a missing API key and how to set it.
    
    Args:
        tool_name: Name of the tool requiring the API key
        env_var_name: Name of the environment variable for the API key
        dotenv_support: Whether the tool supports loading from .env file
    """
    # Escape input values
    tool_name = escape_rich_markup(tool_name)
    env_var_name = escape_rich_markup(env_var_name)
    
    # Determine the operating system
    system = platform.system()
    
    # Create OS-specific instructions for setting the API key
    if system == "Windows":
        env_instructions = f"setx {env_var_name} your_api_key_here"
        env_instructions_temp = f"set {env_var_name}=your_api_key_here"
        env_description = f"[bold green]Option 1: Set environment variable (Windows):[/bold green]\n  • Permanent (new sessions): {env_instructions}\n  • Current session only: {env_instructions_temp}"
    else:  # macOS or Linux
        env_instructions_export = f"export {env_var_name}=your_api_key_here"
        env_instructions_profile = f"echo 'export {env_var_name}=your_api_key_here' >> ~/.bashrc  # or ~/.zshrc"
        env_description = f"[bold green]Option 1: Set environment variable (macOS/Linux):[/bold green]\n  • Current session: {env_instructions_export}\n  • Permanent: {env_instructions_profile}"
    
    if dotenv_support:
        dotenv_instructions = f"Create a .env file in your project directory with:\n  {env_var_name}=your_api_key_here"
        content = f"[bold red]Missing API Key for {tool_name}[/bold red]\n\n[bold white]The {env_var_name} environment variable is not set.[/bold white]\n\n{env_description}\n\n[bold green]Option 2: Use a .env file:[/bold green]\n  {dotenv_instructions}"
    else:
        content = f"[bold red]Missing API Key for {tool_name}[/bold red]\n\n[bold white]The {env_var_name} environment variable is not set.[/bold white]\n\n{env_description}"
    
    # Create and print the panel
    panel = Panel(content, title="[bold yellow]🔑 API Key Required[/bold yellow]", border_style="yellow", expand=False)
    console.print(panel)

def tool_operation(operation: str, result=None):
    """
    Prints a formatted panel for regular tool operations.
    
    Args:
        operation: The operation being performed (e.g., "Adding", "Added", "Removing")
        result: The result of the operation, if available
    """
    table = Table(show_header=False, expand=True, box=None)
    table.width = 60
    
    # Format the operation text
    operation_text = f"[bold magenta]{escape_rich_markup(operation)}[/bold magenta]"
    table.add_row(operation_text)
    
    # If there's a result, add it to the table
    if result:
        result_str = str(result)
        table.add_row("")  # Add spacing
        table.add_row(f"[green]{escape_rich_markup(result_str)}[/green]")
    
    panel = Panel(
        table,
        title="[bold magenta]Upsonic - Tool Operation[/bold magenta]",
        border_style="magenta",
        expand=True,
        width=70
    )
    
    console.print(panel)
    spacing()


from fastapi import HTTPException
from pydantic import BaseModel
from typing import List, Dict, Any, Optional, Union
import traceback
from ...api import app, timeout, handle_server_errors
from ..call import Call
import asyncio
import cloudpickle
cloudpickle.DEFAULT_PROTOCOL = 2
import base64
import os


prefix = "/level_one"


class GPT4ORequest(BaseModel):
    prompt: str
    images: Optional[List[str]] = None
    response_format: Optional[Any] = []
    tools: Optional[Any] = []
    context: Optional[Any] = None
    llm_model: Optional[Any] = "openai/gpt-4o"
    system_prompt: Optional[Any] = None


@app.post(f"{prefix}/gpt4o")
@handle_server_errors
async def call_gpt4o(request: GPT4ORequest):
    """
    Endpoint to call GPT-4 with optional tools and MCP servers.

    Args:
        request: GPT4ORequest containing prompt and optional parameters

    Returns:
        The response from the AI model
    """
    try:
        # Handle pickled response format
        if request.response_format != "str":
            try:
                # Decode and unpickle the response format
                pickled_data = base64.b64decode(request.response_format)
                response_format = cloudpickle.loads(pickled_data)
            except Exception as e:
                tb = traceback.extract_tb(e.__traceback__)
                file_path = tb[-1].filename
                if "Upsonic/src/" in file_path:
                    file_path = file_path.split("Upsonic/src/")[1]
                line_number = tb[-1].lineno
                traceback.print_exc()
                # Fallback to basic type mapping if unpickling fails
                type_mapping = {
                    "str": str,
                    "int": int,
                    "float": float,
                    "bool": bool,
                }
                response_format = type_mapping.get(request.response_format, str)
        else:
            response_format = str

        if request.context is not None:
            try:
                pickled_context = base64.b64decode(request.context)
                context = cloudpickle.loads(pickled_context)
            except Exception as e:
                tb = traceback.extract_tb(e.__traceback__)
                file_path = tb[-1].filename
                if "Upsonic/src/" in file_path:
                    file_path = file_path.split("Upsonic/src/")[1]
                line_number = tb[-1].lineno
                traceback.print_exc()
                context = None
        else:
            context = None

        result = await Call.gpt_4o(
            prompt=request.prompt,
            images=request.images,
            response_format=response_format,
            tools=request.tools,
            context=context,
            llm_model=request.llm_model,
            system_prompt=request.system_prompt
        )

        if request.response_format != "str" and result["status_code"] == 200:
            result["result"] = cloudpickle.dumps(result["result"])
            result["result"] = base64.b64encode(result["result"]).decode('utf-8')
        return {"result": result, "status_code": 200}
    except Exception as e:
        tb = traceback.extract_tb(e.__traceback__)
        file_path = tb[-1].filename
        if "Upsonic/src/" in file_path:
            file_path = file_path.split("Upsonic/src/")[1]
        line_number = tb[-1].lineno
        traceback.print_exc()
        return {"result": {"status_code": 500, "detail": f"Error processing Call request in {file_path} at line {line_number}: {str(e)}"}, "status_code": 500}



from .tasks.tasks import Task
from .direct_llm_call.direct_llm_cal import Direct

from typing import Optional

class Language:
    def __init__(self, language: str, task: Task, llm_model: str):
        self.language = language
        self.task = task
        self.llm_model = llm_model

    async def transform(self):
        language_transformation_task = Task(
            f"User task is completed but we want to change the language of the task to {self.language}. Just return the translated result of task. Dont say or put anything to your return. Make one to one translation.",
            context=[self.task],
            response_format=self.task.response_format,
        )

        direct = Direct(self.llm_model)
        await direct.do_async(language_transformation_task)
        return language_transformation_task.response



import traceback
import anthropic
import openai
from pydantic import BaseModel
import os
from pydantic_ai.messages import ImageUrl

from typing import Any, Optional, List

from ...storage.configuration import Configuration

from ..level_utilized.memory import save_temporary_memory, get_temporary_memory

from ..level_utilized.utility import (
    agent_creator, 
    prepare_message_history,
    process_error_traceback,
    format_response,
    handle_compression_retry
)

from ...client.tasks.tasks import Task
from ...client.tasks.task_response import ObjectResponse

from ..level_one.call import Call


def extract_latest_tool_usage(all_messages):
    """Extract tool usage from the latest interaction only."""
    tool_usage = []
    current_tool = None
    
    # Find the start of the latest interaction
    latest_interaction_start = 0
    for i, msg in enumerate(all_messages):
        if msg.kind == 'request' and any(part.part_kind == 'user-prompt' for part in msg.parts):
            latest_interaction_start = i
            
    # Process only messages from the latest interaction
    for msg in all_messages[latest_interaction_start:]:
        if msg.kind == 'request':
            for part in msg.parts:
                if part.part_kind == 'tool-return':
                    if current_tool and current_tool['tool_name'] != 'final_result':
                        current_tool['tool_result'] = part.content
                        tool_usage.append(current_tool)
                    current_tool = None
                    
        elif msg.kind == 'response':
            for part in msg.parts:
                if part.part_kind == 'tool-call' and part.tool_name != 'final_result':
                    current_tool = {
                        'tool_name': part.tool_name,
                        'params': part.args,
                        'tool_result': None
                    }
    
    return tool_usage

class AgentManager:
    async def agent(
        self,
        agent_id: str,
        prompt: str,
        images: Optional[List[str]] = None,
        response_format: BaseModel = str,
        tools: list[str] = [],
        context: Any = None,
        llm_model: str = "openai/gpt-4o",
        system_prompt: Optional[Any] = None,
        context_compress: bool = False,
        memory: bool = False
    ):
        try:
            roulette_agent = agent_creator(
                response_format=response_format, 
                tools=tools, 
                context=context, 
                llm_model=llm_model, 
                system_prompt=system_prompt,
                context_compress=context_compress
            )
            
            if isinstance(roulette_agent, dict) and "status_code" in roulette_agent:
                return roulette_agent  # Return error from agent_creator

            agent_memory = []
            if memory:
                agent_memory = get_temporary_memory(agent_id)

            message_history = prepare_message_history(prompt, images, llm_model, tools)
                
            total_request_tokens = 0
            total_response_tokens = 0

            try:
                result = await roulette_agent.run(message_history, message_history=agent_memory)
            except (openai.BadRequestError, anthropic.BadRequestError) as e:
                str_e = str(e)
                if "400" in str_e and context_compress:
                    try:
                        result = await handle_compression_retry(
                            prompt, images, tools, llm_model,
                            response_format, context, system_prompt, agent_memory
                        )
                    except Exception as e:
                        return process_error_traceback(e)
                else:
                    return process_error_traceback(e)

            total_request_tokens += result.usage().request_tokens
            total_response_tokens += result.usage().response_tokens

            if memory:
                save_temporary_memory(result.all_messages(), agent_id)

            # Extract tool usage from the latest interaction only
            tool_usage = extract_latest_tool_usage(result.all_messages())

            return {
                "status_code": 200, 
                "result": result.data, 
                "usage": {
                    "input_tokens": total_request_tokens, 
                    "output_tokens": total_response_tokens
                },
                "tool_usage": tool_usage
            }

        except Exception as e:
            return process_error_traceback(e)


Agent = AgentManager()


from ..agent_configuration.agent_configuration import get_or_create_client, register_tools
from ..tasks.tasks import Task
from typing import Any, Callable, TypeVar, cast

T = TypeVar('T')

from ...model_registry import ModelNames

class DirectStatic:
    """Static methods for making direct LLM calls using the Upsonic client."""
    
    @staticmethod
    def do(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Execute a direct LLM call with the given task and model.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
            
        Returns:
            The response from the LLM
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    DirectStatic.do_async(task, model, client, debug, retry), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(DirectStatic.do_async(task, model, client, debug, retry))

    @staticmethod
    async def do_async(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Execute a direct LLM call with the given task and model asynchronously.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
            
        Returns:
            The response from the LLM
        """
        global latest_upsonic_client
        from ..latest_upsonic_client import latest_upsonic_client

        # Use provided client or get/create one
        if client is not None:
            the_client = client
        else:
            the_client = get_or_create_client(debug=debug)
        
        # Register tools if needed
        the_client = register_tools(the_client, task.tools)

        # Execute the direct call asynchronously with retry parameter
        await the_client.call_async(task, model, retry=retry)
        return task.response

    @staticmethod
    def print_do(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Execute a direct LLM call and print the result.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
            
        Returns:
            The response from the LLM
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    DirectStatic.print_do_async(task, model, client, debug, retry), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(DirectStatic.print_do_async(task, model, client, debug, retry))

    @staticmethod
    async def print_do_async(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Execute a direct LLM call and print the result asynchronously.
        
        Args:
            task: The task to execute
            model: The LLM model to use (default: "openai/gpt-4")
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
            
        Returns:
            The response from the LLM
        """
        result = await DirectStatic.do_async(task, model, client, debug, retry)
        print(result)
        return result


class DirectInstance:
    """Instance-based class for making direct LLM calls using the Upsonic client."""
    
    def __init__(self, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Initialize a DirectInstance with specific model and client settings.
        
        Args:
            model: The LLM model to use (default: None)
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
        """
        self.model = model
        self.client = client
        self.debug = debug
        self.retry = retry
    
    def do(self, task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int | None = None):
        """
        Execute a direct LLM call using instance defaults or overrides.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            retry: Number of retries for failed calls (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    self.do_async(task, model, client, debug, retry), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.do_async(task, model, client, debug, retry))

    async def do_async(self, task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int | None = None):
        """
        Execute a direct LLM call using instance defaults or overrides asynchronously.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            retry: Number of retries for failed calls (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        # Use provided parameters or instance defaults
        actual_model = model if model is not None else self.model
        actual_client = client if client is not None else self.client
        actual_debug = debug if debug is not False else self.debug
        actual_retry = retry if retry is not None else self.retry
        
        # Call the static method with the resolved parameters
        return await DirectStatic.do_async(task, actual_model, actual_client, actual_debug, actual_retry)
        
    def print_do(self, task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int | None = None):
        """
        Execute a direct LLM call and print the result.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            retry: Number of retries for failed calls (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        import asyncio
        
        try:
            # Check if there's a running event loop
            loop = asyncio.get_running_loop()
            if loop.is_running():
                # If there's a running loop, run the coroutine in that loop
                return asyncio.run_coroutine_threadsafe(
                    self.print_do_async(task, model, client, debug, retry), 
                    loop
                ).result()
        except RuntimeError:
            # No running event loop
            pass
        
        # If no running loop or exception occurred, create a new one
        return asyncio.run(self.print_do_async(task, model, client, debug, retry))

    async def print_do_async(self, task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int | None = None):
        """
        Execute a direct LLM call and print the result asynchronously.
        
        Args:
            task: The task to execute
            model: The LLM model to use (overrides instance default if provided)
            client: Optional custom client (overrides instance default if provided)
            debug: Whether to enable debug mode (overrides instance default if provided)
            retry: Number of retries for failed calls (overrides instance default if provided)
            
        Returns:
            The response from the LLM
        """
        result = await self.do_async(task, model, client, debug, retry)
        print(result)
        return result


class Direct:
    """
    Router class that provides both static and instance-based approaches for direct LLM calls.
    
    When used without initialization, it provides static methods.
    When initialized with parameters, it returns an instance-based object.
    
    Example:
        # Correct usage with named model parameter:
        direct = Direct(model="openai/gpt-4o")
        direct = Direct(model="claude/claude-3-5-sonnet")
        
        # Incorrect usage:
        # direct = Direct("openai/gpt-4o")  # Wrong! Must use model=
        # direct = Direct("Researcher Direct")  # Wrong!
        
        # For agent-based operations, use Agent instead:
        # agent = Agent("Researcher Agent")  # Correct!
    """
    
    # Static methods that delegate to DirectStatic
    @staticmethod
    def do(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        return DirectStatic.do(task, model, client, debug, retry)
    
    @staticmethod
    def print_do(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        return DirectStatic.print_do(task, model, client, debug, retry)

    @staticmethod
    async def do_async(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        return await DirectStatic.do_async(task, model, client, debug, retry)
    
    @staticmethod
    async def print_do_async(task: Task, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        return await DirectStatic.print_do_async(task, model, client, debug, retry)
    
    def __new__(cls, *args, model: ModelNames | None = None, client: Any = None, debug: bool = False, retry: int = 3):
        """
        Factory method that returns a DirectInstance object when initialized.
        
        Args:
            model: The LLM model to use (default: None)
            client: Optional custom client to use instead of creating a new one
            debug: Whether to enable debug mode
            retry: Number of retries for failed calls (default: 3)
            
        Returns:
            A DirectInstance object
            
        Raises:
            ValueError: If positional arguments are provided instead of using the named parameter 'model'
        """
        if args:
            raise ValueError(
                "Direct() does not accept positional arguments. Use named parameter 'model' instead.\n"
                "Example: Direct(model='openai/gpt-4o') instead of Direct('openai/gpt-4o')"
            )
            
        return DirectInstance(model, client, debug, retry)



from typing import List, Dict, Any, Optional, Union, Callable, Set
from pydantic import BaseModel, Field
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.markup import escape
from rich.progress import Progress, TextColumn, BarColumn, SpinnerColumn, TimeElapsedColumn, TimeRemainingColumn
import uuid
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

from .printing import console, spacing, escape_rich_markup
from .tasks.tasks import Task
from .tasks.task_response import ObjectResponse
from .agent_configuration.agent_configuration import AgentConfiguration

# Define DecisionResponse at module level
class DecisionResponse(ObjectResponse):
    """Response type for LLM-based decisions that returns a boolean result."""
    result: bool

# Import Direct for type checking
try:
    from .direct_llm.direct import Direct
except ImportError:
    # Define a placeholder for type checking if the import fails
    class Direct:
        pass


class DecisionLLM(BaseModel):
    """
    A decision node that uses a language model to evaluate input and determine execution flow.
    
    Attributes:
        description: Human-readable description of the decision
        true_branch: The branch to follow if the LLM decides yes/true
        false_branch: The branch to follow if the LLM decides no/false
        id: Unique identifier for this decision node
    """
    description: str
    true_branch: Optional[Union['TaskNode', 'TaskChain', 'DecisionFunc', 'DecisionLLM']] = None
    false_branch: Optional[Union['TaskNode', 'TaskChain', 'DecisionFunc', 'DecisionLLM']] = None
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    
    class Config:
        arbitrary_types_allowed = True
    
    def __init__(self, description: str, *, true_branch=None, false_branch=None, id=None, **kwargs):
        """
        Initialize a DecisionLLM with a positional description parameter.
        
        Args:
            description: Human-readable description of the decision
            true_branch: The branch to follow if the LLM decides yes/true
            false_branch: The branch to follow if the LLM decides no/false
            id: Unique identifier for this decision node
        """
        if id is None:
            id = str(uuid.uuid4())
        super().__init__(description=description, true_branch=true_branch, false_branch=false_branch, id=id, **kwargs)
    
    def evaluate(self, data: Any) -> bool:
        """
        Evaluates the decision using an LLM with the provided data.
        
        This is a placeholder that will be replaced during graph execution with
        actual LLM inference using the graph's default agent.
        
        Args:
            data: Data to evaluate (typically the output of the previous task)
            
        Returns:
            True if the LLM determines yes/true, False otherwise
        """
        # This is a placeholder - the actual implementation happens 
        # during graph execution using the graph's agent
        return True
    
    def _generate_prompt(self, data: Any) -> str:
        """
        Generates a prompt for the LLM based on the decision description and input data.
        
        Args:
            data: The data to be evaluated (typically the output of the previous task)
            
        Returns:
            A formatted prompt string for the LLM
        """
        prompt = f"""
You are an decision node in a graph.

Decision question: {self.description}

Previous node output:
<data>
{data}
</data>
"""
        return prompt.strip()
    
    def if_true(self, branch: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'DecisionLLM':
        """
        Sets the branch to follow if the LLM evaluates to True/Yes.
        
        Args:
            branch: The node, task, or chain to execute if the LLM decides yes/true
            
        Returns:
            Self for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(branch, Task):
            branch = TaskNode(task=branch)
            
        self.true_branch = branch
        return self
    
    def if_false(self, branch: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'DecisionLLM':
        """
        Sets the branch to follow if the LLM evaluates to False/No.
        
        Args:
            branch: The node, task, or chain to execute if the LLM decides no/false
            
        Returns:
            Self for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(branch, Task):
            branch = TaskNode(task=branch)
            
        self.false_branch = branch
        return self
    
    def __rshift__(self, other: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'TaskChain':
        """
        Implements the >> operator to chain this decision with another node.
        Both the true and false branches will converge to the specified node.
        
        Args:
            other: The node, task, or chain to connect after both branches
            
        Returns:
            A TaskChain containing the decision and its branches
        """
        chain = TaskChain()
        chain.nodes.append(self)
        
        # Add the next node/chain
        if isinstance(other, Task):
            other_node = TaskNode(task=other)
            chain.nodes.append(other_node)
        elif isinstance(other, TaskNode):
            chain.nodes.append(other)
        elif isinstance(other, TaskChain):
            chain.nodes.extend(other.nodes)
        elif isinstance(other, (DecisionFunc, DecisionLLM)):
            chain.nodes.append(other)
            
        return chain


class DecisionFunc(BaseModel):
    """
    A decision node that evaluates a condition function on task output to determine execution flow.
    
    Attributes:
        description: Human-readable description of the decision
        func: The function that evaluates the condition
        true_branch: The branch to follow if the condition is true
        false_branch: The branch to follow if the condition is false
        id: Unique identifier for this decision node
    """
    description: str
    func: Callable
    true_branch: Optional[Union['TaskNode', 'TaskChain', 'DecisionFunc', 'DecisionLLM']] = None
    false_branch: Optional[Union['TaskNode', 'TaskChain', 'DecisionFunc', 'DecisionLLM']] = None
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    
    class Config:
        arbitrary_types_allowed = True
    
    def __init__(self, description: str, func: Callable, *, true_branch=None, false_branch=None, id=None, **kwargs):
        """
        Initialize a DecisionFunc with positional description and func parameters.
        
        Args:
            description: Human-readable description of the decision
            func: The function that evaluates the condition
            true_branch: The branch to follow if the condition is true
            false_branch: The branch to follow if the condition is false
            id: Unique identifier for this decision node
        """
        if id is None:
            id = str(uuid.uuid4())
        super().__init__(description=description, func=func, true_branch=true_branch, false_branch=false_branch, id=id, **kwargs)
        
    def evaluate(self, data: Any) -> bool:
        """
        Evaluates the condition function with the provided data.
        
        Args:
            data: Data to evaluate (typically the output of the previous task)
            
        Returns:
            True if condition passes, False otherwise
        """
        return self.func(data)
    
    def if_true(self, branch: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'DecisionFunc':
        """
        Sets the branch to follow if the condition evaluates to True.
        
        Args:
            branch: The node, task, or chain to execute if condition is true
            
        Returns:
            Self for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(branch, Task):
            branch = TaskNode(task=branch)
            
        self.true_branch = branch
        return self
    
    def if_false(self, branch: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'DecisionFunc':
        """
        Sets the branch to follow if the condition evaluates to False.
        
        Args:
            branch: The node, task, or chain to execute if condition is false
            
        Returns:
            Self for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(branch, Task):
            branch = TaskNode(task=branch)
            
        self.false_branch = branch
        return self
    
    def __rshift__(self, other: Union['TaskNode', Task, 'TaskChain', 'DecisionFunc', 'DecisionLLM']) -> 'TaskChain':
        """
        Implements the >> operator to chain this decision with another node.
        Both the true and false branches will converge to the specified node.
        
        Args:
            other: The node, task, or chain to connect after both branches
            
        Returns:
            A TaskChain containing the decision and its branches
        """
        chain = TaskChain()
        chain.nodes.append(self)
        
        # Add the next node/chain
        if isinstance(other, Task):
            other_node = TaskNode(task=other)
            chain.nodes.append(other_node)
        elif isinstance(other, TaskNode):
            chain.nodes.append(other)
        elif isinstance(other, TaskChain):
            chain.nodes.extend(other.nodes)
        elif isinstance(other, (DecisionFunc, DecisionLLM)):
            chain.nodes.append(other)
            
        return chain


class TaskNode(BaseModel):
    """
    Wrapper around a Task that adds graph connectivity features.
    
    Attributes:
        task: The Task object this node wraps
        id: Unique identifier for this node
    """
    task: Task
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    
    # For edge management
    next_nodes: List['TaskNode'] = Field(default_factory=list)
    
    def __rshift__(self, other: Union['TaskNode', Task, 'DecisionFunc', 'DecisionLLM']) -> 'TaskChain':
        """
        Implements the >> operator to connect nodes in a chain.
        
        Args:
            other: The next node or task in the chain
            
        Returns:
            A TaskChain object containing both nodes
        """
        chain = TaskChain()
        chain.add(self)
        
        # If the other object is a Task, wrap it in a TaskNode
        if isinstance(other, Task):
            other = TaskNode(task=other)
            
        chain.add(other)
        return chain


class TaskChain:
    """
    Represents a chain of connected task nodes.
    
    Attributes:
        nodes: List of nodes in the chain
        edges: Dictionary mapping node IDs to their next nodes
    """
    def __init__(self):
        self.nodes: List[Union[TaskNode, DecisionFunc, DecisionLLM]] = []
        self.edges: Dict[str, List[str]] = {}
        
    def add(self, node: Union[TaskNode, Task, 'TaskChain', DecisionFunc, DecisionLLM]) -> 'TaskChain':
        """
        Adds a node or another chain to this chain.
        
        Args:
            node: The node, task, or chain to add
            
        Returns:
            This chain for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(node, Task):
            node = TaskNode(task=node)
        
        if isinstance(node, TaskNode):
            if self.nodes:
                last_node = self.nodes[-1]
                if last_node.id not in self.edges:
                    self.edges[last_node.id] = []
                self.edges[last_node.id].append(node.id)
                
                # If the last node is a TaskNode, update its next_nodes
                if isinstance(last_node, TaskNode):
                    last_node.next_nodes.append(node)
            self.nodes.append(node)
        elif isinstance(node, (DecisionFunc, DecisionLLM)):
            # Add the decision node itself
            self.nodes.append(node)
            
            # Connect decision node to its branches
            if node.true_branch:
                if isinstance(node.true_branch, TaskNode):
                    if node.id not in self.edges:
                        self.edges[node.id] = []
                    self.edges[node.id].append(node.true_branch.id)
                    
                    # Add true branch node if not already in the chain
                    if node.true_branch not in self.nodes:
                        self.nodes.append(node.true_branch)
                elif isinstance(node.true_branch, TaskChain):
                    # Add all nodes from the true branch chain
                    if node.true_branch.nodes:
                        first_node = node.true_branch.nodes[0]
                        if node.id not in self.edges:
                            self.edges[node.id] = []
                        self.edges[node.id].append(first_node.id)
                        
                        # Add the nodes from the true branch
                        for n in node.true_branch.nodes:
                            if n not in self.nodes:
                                self.nodes.append(n)
                        
                        # Merge the edges from the true branch
                        for src, targets in node.true_branch.edges.items():
                            if src not in self.edges:
                                self.edges[src] = []
                            self.edges[src].extend(targets)
            
            # Connect false branch as well
            if node.false_branch:
                if isinstance(node.false_branch, TaskNode):
                    if node.id not in self.edges:
                        self.edges[node.id] = []
                    self.edges[node.id].append(node.false_branch.id)
                    
                    # Add false branch node if not already in the chain
                    if node.false_branch not in self.nodes:
                        self.nodes.append(node.false_branch)
                elif isinstance(node.false_branch, TaskChain):
                    # Add all nodes from the false branch chain
                    if node.false_branch.nodes:
                        first_node = node.false_branch.nodes[0]
                        if node.id not in self.edges:
                            self.edges[node.id] = []
                        self.edges[node.id].append(first_node.id)
                        
                        # Add the nodes from the false branch
                        for n in node.false_branch.nodes:
                            if n not in self.nodes:
                                self.nodes.append(n)
                        
                        # Merge the edges from the false branch
                        for src, targets in node.false_branch.edges.items():
                            if src not in self.edges:
                                self.edges[src] = []
                            self.edges[src].extend(targets)
        elif isinstance(node, TaskChain):
            if self.nodes and node.nodes:
                last_node = self.nodes[-1]
                first_of_new = node.nodes[0]
                if last_node.id not in self.edges:
                    self.edges[last_node.id] = []
                self.edges[last_node.id].append(first_of_new.id)
                
                # If the last node is a TaskNode, update its next_nodes
                if isinstance(last_node, TaskNode) and isinstance(first_of_new, TaskNode):
                    last_node.next_nodes.append(first_of_new)
            
            # Merge the other chain's edges into this one
            self.edges.update(node.edges)
            self.nodes.extend(node.nodes)
            
        return self
        
    def __rshift__(self, other: Union[TaskNode, Task, 'TaskChain', DecisionFunc, DecisionLLM]) -> 'TaskChain':
        """
        Implements the >> operator to connect this chain with another node, task, or chain.
        
        Args:
            other: The next node, task, or chain to connect
            
        Returns:
            This chain with the new node(s) added
        """
        return self.add(other)


class State(BaseModel):
    """
    Manages the state between task executions in the graph.
    
    Attributes:
        data: Dictionary storing additional data shared across tasks
        task_outputs: Dictionary mapping node IDs to their task outputs
    """
    data: Dict[str, Any] = Field(default_factory=dict)
    task_outputs: Dict[str, Any] = Field(default_factory=dict)
    
    def update(self, node_id: str, output: Any):
        """
        Updates the state with a node's task output.
        
        Args:
            node_id: ID of the node
            output: Output from the task execution
        """
        self.task_outputs[node_id] = output
        
    def get_task_output(self, node_id: str) -> Any:
        """
        Retrieves the output of a specific node's task.
        
        Args:
            node_id: ID of the node
            
        Returns:
            The output of the specified node's task
        """
        return self.task_outputs.get(node_id)
    
    def get_latest_output(self) -> Any:
        """
        Gets the most recent task output.
        
        Returns:
            The output of the most recently executed task
        """
        if not self.task_outputs:
            return None
        
        # Return the most recently added output
        return list(self.task_outputs.values())[-1]


class Graph(BaseModel):
    """
    Main graph structure that manages task execution, state, and workflow.
    
    Attributes:
        default_agent: Default agent to use when a task doesn't specify one
        parallel_execution: Whether to execute independent tasks in parallel
        max_parallel_tasks: Maximum number of tasks to execute in parallel
        show_progress: Whether to display a progress bar during execution
    """
    # Accept either AgentConfiguration or Direct as the default_agent
    default_agent: Optional[Any] = None
    parallel_execution: bool = False
    max_parallel_tasks: int = 4
    show_progress: bool = True
    
    # Private attributes (not part of the model schema)
    nodes: List[Union[TaskNode, DecisionFunc, DecisionLLM]] = Field(default_factory=list)
    edges: Dict[str, List[str]] = Field(default_factory=dict)
    state: State = Field(default_factory=State)
    
    class Config:
        arbitrary_types_allowed = True
    
    def __init__(self, **data):
        # Validate that default_agent is either AgentConfiguration or Direct
        if 'default_agent' in data and data['default_agent'] is not None:
            agent = data['default_agent']
            # Check if it has the 'do' method which both types should have
            if not hasattr(agent, 'do') or not callable(getattr(agent, 'do')):
                raise ValueError("default_agent must be an instance of AgentConfiguration or Direct with a 'do' method")
        super().__init__(**data)
    
    def add(self, tasks_chain: Union[Task, TaskNode, TaskChain, DecisionFunc, DecisionLLM]) -> 'Graph':
        """
        Adds tasks to the graph.
        
        Args:
            tasks_chain: A Task, TaskNode, TaskChain, DecisionFunc, or DecisionLLM to add to the graph
            
        Returns:
            This graph for method chaining
        """
        # If given a Task, wrap it in a TaskNode
        if isinstance(tasks_chain, Task):
            tasks_chain = TaskNode(task=tasks_chain)
        
        if isinstance(tasks_chain, TaskNode):
            self.nodes.append(tasks_chain)
        elif isinstance(tasks_chain, (DecisionFunc, DecisionLLM)):
            self.nodes.append(tasks_chain)
            
            # Handle decision branches
            if tasks_chain.true_branch:
                if isinstance(tasks_chain.true_branch, TaskNode):
                    if tasks_chain.true_branch not in self.nodes:
                        self.nodes.append(tasks_chain.true_branch)
                    if tasks_chain.id not in self.edges:
                        self.edges[tasks_chain.id] = []
                    self.edges[tasks_chain.id].append(tasks_chain.true_branch.id)
                elif isinstance(tasks_chain.true_branch, TaskChain):
                    # Add all nodes and edges from the true branch chain
                    for node in tasks_chain.true_branch.nodes:
                        if node not in self.nodes:
                            self.nodes.append(node)
                    
                    # Add the connection from decision to first node of true branch
                    if tasks_chain.true_branch.nodes:
                        first_node = tasks_chain.true_branch.nodes[0]
                        if tasks_chain.id not in self.edges:
                            self.edges[tasks_chain.id] = []
                        self.edges[tasks_chain.id].append(first_node.id)
                    
                    # Merge edges from the true branch
                    for src, targets in tasks_chain.true_branch.edges.items():
                        if src not in self.edges:
                            self.edges[src] = []
                        self.edges[src].extend(targets)
            
            # Handle false branch
            if tasks_chain.false_branch:
                if isinstance(tasks_chain.false_branch, TaskNode):
                    if tasks_chain.false_branch not in self.nodes:
                        self.nodes.append(tasks_chain.false_branch)
                    if tasks_chain.id not in self.edges:
                        self.edges[tasks_chain.id] = []
                    self.edges[tasks_chain.id].append(tasks_chain.false_branch.id)
                elif isinstance(tasks_chain.false_branch, TaskChain):
                    # Add all nodes and edges from the false branch chain
                    for node in tasks_chain.false_branch.nodes:
                        if node not in self.nodes:
                            self.nodes.append(node)
                    
                    # Add the connection from decision to first node of false branch
                    if tasks_chain.false_branch.nodes:
                        first_node = tasks_chain.false_branch.nodes[0]
                        if tasks_chain.id not in self.edges:
                            self.edges[tasks_chain.id] = []
                        self.edges[tasks_chain.id].append(first_node.id)
                    
                    # Merge edges from the false branch
                    for src, targets in tasks_chain.false_branch.edges.items():
                        if src not in self.edges:
                            self.edges[src] = []
                        self.edges[src].extend(targets)
                    
        elif isinstance(tasks_chain, TaskChain):
            self.nodes.extend(tasks_chain.nodes)
            self.edges.update(tasks_chain.edges)
            
        return self
    
    def _get_available_agent(self) -> Any:
        """
        Finds an available agent either from the graph default or from any task node.
        
        Returns:
            An agent that can be used for execution, or None if none is found
        """
        # First check if we have a default agent
        if self.default_agent is not None:
            return self.default_agent
        
        # If no default agent, check all task nodes for an agent
        for node in self.nodes:
            if isinstance(node, TaskNode) and node.task.agent is not None:
                return node.task.agent
        
        # No agent found
        return None

    def _execute_task(self, node: TaskNode, state: State, verbose: bool = False) -> Any:
        """
        Executes a single task.
        
        Args:
            node: The TaskNode containing the task to execute
            state: Current state object
            verbose: Whether to print detailed information
            
        Returns:
            The output of the task
        """
        task = node.task
        
        # Use the task's agent or try to find an available agent
        runner = task.agent or self.default_agent
        if runner is None:
            # Try to find any agent from other task nodes
            runner = self._get_available_agent()
            if runner is None:
                raise ValueError(f"No agent specified for task '{task.description}' and no default agent set")
        
        try:
            # Start timing
            start_time = time.time()
            task.start_time = start_time
            
            if verbose:
                # Create and print a task execution panel
                table = Table(show_header=False, expand=True, box=None)
                table.add_row("[bold]Task:[/bold]", f"[cyan]{escape_rich_markup(task.description)}[/cyan]")
                # Display runner type safely
                runner_type = runner.__class__.__name__ if hasattr(runner, '__class__') else type(runner).__name__
                table.add_row("[bold]Agent:[/bold]", f"[yellow]{escape_rich_markup(runner_type)}[/yellow]")
                if task.tools:
                    tool_names = [escape_rich_markup(t.__class__.__name__ if hasattr(t, '__class__') else str(t)) for t in task.tools]
                    table.add_row("[bold]Tools:[/bold]", f"[green]{escape_rich_markup(', '.join(tool_names))}[/green]")
                panel = Panel(
                    table,
                    title="[bold blue]Upsonic - Executing Task[/bold blue]",
                    border_style="blue",
                    expand=True,
                    width=70
                )
                console.print(panel)
                spacing()
            
            # Get previous outputs if available
            previous_outputs = [state.get_task_output(prev_node.id) for prev_node in self._get_predecessors(node)]
            
            # Add previous outputs as context if appropriate
            if previous_outputs and not task.context:
                task.context = previous_outputs
            
            # Execute the task - both AgentConfiguration and Direct have the do method
            output = runner.do(task)
            
            # End timing
            end_time = time.time()
            task.end_time = end_time
            
            if verbose:
                # Create and print a task completion panel
                time_taken = end_time - start_time
                table = Table(show_header=False, expand=True, box=None)
                table.add_row("[bold]Task:[/bold]", f"[cyan]{escape_rich_markup(task.description)}[/cyan]")
                
                # Handle different output types for display
                output_str = self._format_output_for_display(output)
                
                table.add_row("[bold]Output:[/bold]", f"[green]{output_str}[/green]")
                table.add_row("[bold]Time Taken:[/bold]", f"{time_taken:.2f} seconds")
                if task.total_cost:
                    table.add_row("[bold]Estimated Cost:[/bold]", f"${task.total_cost:.4f}")
                panel = Panel(
                    table,
                    title="[bold green]✅ Task Completed[/bold green]",
                    border_style="green",
                    expand=True,
                    width=70
                )
                console.print(panel)
                spacing()
            
            return output
            
        except Exception as e:
            if verbose:
                console.print(f"[bold red]Task '{escape_rich_markup(task.description)}' failed: {escape_rich_markup(str(e))}[/bold red]")
            raise
    
    def _evaluate_decision(self, decision_node: Union[DecisionFunc, DecisionLLM], state: State, verbose: bool = False) -> Union[TaskNode, TaskChain, None]:
        """
        Evaluates a decision node to determine which branch to follow.
        
        Args:
            decision_node: The decision node to evaluate
            state: Current state object
            verbose: Whether to print detailed information
            
        Returns:
            The branch to follow (true or false)
        """
        # Get the most recent output to evaluate
        latest_output = state.get_latest_output()
        
        # Evaluate differently based on the decision node type
        if isinstance(decision_node, DecisionFunc):
            # For function-based decisions, directly call the function
            result = decision_node.evaluate(latest_output)
        elif isinstance(decision_node, DecisionLLM):
            # For LLM-based decisions, use the default agent or find an available one
            agent = self.default_agent
            if agent is None:
                # Try to find any agent from task nodes
                agent = self._get_available_agent()
                if agent is None:
                    raise ValueError(f"No agent available for LLM-based decision: '{decision_node.description}'")
            
            # Generate the prompt for the LLM
            prompt = decision_node._generate_prompt(latest_output)
            
            # Create a temporary task for the LLM to execute
            decision_task = Task(prompt,
                response_format=DecisionResponse
            )
            
            # Execute the task using the agent
            response = agent.do(decision_task)
            
            # Get the boolean result from the structured response
            result = response.result if hasattr(response, 'result') else False
            
            if verbose:
                console.print(f"[dim]LLM Decision Response: {escape_rich_markup(str(response))}[/dim]")
                console.print(f"[dim]Decision Result: {'Yes' if result else 'No'}[/dim]")
        else:
            raise ValueError(f"Unknown decision node type: {type(decision_node)}")
        
        if verbose:
            # Create and print a decision evaluation panel
            table = Table(show_header=False, expand=True, box=None)
            table.add_row("[bold]Decision:[/bold]", f"[cyan]{escape_rich_markup(decision_node.description)}[/cyan]")
            table.add_row("[bold]Result:[/bold]", f"[green]{result}[/green]")
            panel = Panel(
                table,
                title="[bold yellow]🔀 Evaluating Decision[/bold yellow]",
                border_style="yellow",
                expand=True,
                width=70
            )
            console.print(panel)
            spacing()
        
        # Return the appropriate branch
        if result:
            return decision_node.true_branch
        else:
            return decision_node.false_branch
    
    def _format_output_for_display(self, output: Any) -> str:
        """
        Format an output value for display in verbose mode.
        
        Args:
            output: The output value to format
            
        Returns:
            A string representation of the output
        """
        # If output is None, return empty string
        if output is None:
            return ""
        
        # If output is a Pydantic model
        if hasattr(output, '__class__') and hasattr(output.__class__, 'model_dump'):
            try:
                # Try to get a compact representation of the model
                model_dict = output.model_dump()
                # Format as compact JSON with max 200 chars
                import json
                output_str = json.dumps(model_dict, default=str)
                if len(output_str) > 200:
                    output_str = output_str[:197] + "..."
                return escape_rich_markup(output_str)
            except Exception:
                # Fallback to str if model_dump fails
                output_str = str(output)
        else:
            # Regular string representation
            output_str = str(output)
        
        # Truncate if too long
        if len(output_str) > 200:
            output_str = output_str[:197] + "..."
            
        return escape_rich_markup(output_str)
    
    def _get_predecessors(self, node: Union[TaskNode, DecisionFunc, DecisionLLM]) -> List[Union[TaskNode, DecisionFunc, DecisionLLM]]:
        """
        Gets the predecessor nodes that feed into the given node.
        
        Args:
            node: The node to find predecessors for
            
        Returns:
            List of predecessor nodes
        """
        predecessors = []
        # Check node connections in the edges dictionary
        for n_id, next_ids in self.edges.items():
            if node.id in next_ids:
                # Find the node object
                for n in self.nodes:
                    if n.id == n_id:
                        predecessors.append(n)
                        break
        
        # Also check TaskNode's next_nodes connections
        for n in self.nodes:
            if isinstance(n, TaskNode):
                for next_node in n.next_nodes:
                    if next_node.id == node.id:
                        predecessors.append(n)
        
        return predecessors
    
    def _get_start_nodes(self) -> List[Union[TaskNode, DecisionFunc, DecisionLLM]]:
        """
        Gets the starting nodes of the graph (those with no predecessors).
        
        Returns:
            List of start nodes
        """
        # Get all IDs that appear as targets in the edges dictionary
        all_target_ids = {target_id for targets in self.edges.values() for target_id in targets}
        
        # Also get all IDs that appear as targets in TaskNode.next_nodes
        for n in self.nodes:
            if isinstance(n, TaskNode):
                for next_node in n.next_nodes:
                    all_target_ids.add(next_node.id)
        
        # Return nodes that don't appear as targets
        return [node for node in self.nodes if node.id not in all_target_ids]
    
    def _get_next_nodes(self, node: Union[TaskNode, DecisionFunc, DecisionLLM]) -> List[Union[TaskNode, DecisionFunc, DecisionLLM]]:
        """
        Gets the nodes that come after the given node.
        
        Args:
            node: The node to find successors for
            
        Returns:
            List of successor nodes
        """
        next_nodes = []
        
        # Check edges dictionary
        if node.id in self.edges:
            next_ids = self.edges[node.id]
            for next_id in next_ids:
                for n in self.nodes:
                    if n.id == next_id:
                        next_nodes.append(n)
        
        # If it's a TaskNode, also check its next_nodes attribute
        if isinstance(node, TaskNode):
            for next_node in node.next_nodes:
                if next_node not in next_nodes:  # Avoid duplicates
                    next_nodes.append(next_node)
        
        return next_nodes
    
    def _run_sequential(self, verbose: bool = False, show_progress: bool = True) -> State:
        """
        Runs tasks sequentially with support for decision nodes.
        
        Args:
            verbose: Whether to print detailed information
            show_progress: Whether to display a progress bar
            
        Returns:
            The final state object
        """
        if verbose:
            console.print(f"[blue]Executing graph with decision support[/blue]")
            spacing()
        
        # We can't use topological sort with decisions, so we use a dynamic execution approach
        execution_queue = self._get_start_nodes()
        executed_nodes = set()
        
        # Count all possible nodes in the graph including all branches
        all_nodes = self._count_all_possible_nodes()
        
        if show_progress:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                TimeElapsedColumn(),
                console=console
            ) as progress:
                # Set total to all possible nodes
                overall_task = progress.add_task("[bold blue]Graph Execution", total=all_nodes)
                
                # Process nodes until the queue is empty
                completed_nodes = 0
                while execution_queue:
                    node = execution_queue.pop(0)
                    
                    # Update progress description
                    if isinstance(node, TaskNode):
                        desc = f"[bold blue]Graph Execution - Current: [cyan]{escape_rich_markup(node.task.description[:40])}{'...' if len(node.task.description) > 40 else ''}[/cyan]"
                    else:
                        desc = f"[bold blue]Graph Execution - Evaluating Decision: [yellow]{escape_rich_markup(node.description)}[/yellow]"
                    
                    # Update the progress bar
                    progress.update(overall_task, description=desc)
                    
                    # Process the node
                    if isinstance(node, TaskNode):
                        # Get the latest output for debugging
                        if verbose:
                            latest_output = self.state.get_latest_output()
                            if latest_output:
                                console.print(f"[dim]********latest_output[/dim]")
                                console.print(f"[dim]{escape_rich_markup(str(latest_output))}[/dim]")
                                
                        # If this task doesn't have context but there are previous outputs, add them as context
                        if not node.task.context and self.state.task_outputs:
                            # Get the latest output to use as context
                            latest_output = self.state.get_latest_output()
                            if latest_output:
                                # Set the context for this task
                                node.task.context = [latest_output]
                                if verbose:
                                    console.print(f"[dim]Setting context from previous output for task: {escape_rich_markup(node.task.description)}[/dim]")
                        
                        output = self._execute_task(node, self.state, verbose)
                        self.state.update(node.id, output)
                        executed_nodes.add(node.id)
                        
                        # Add successor nodes to the queue
                        for next_node in self._get_next_nodes(node):
                            # Only add if all predecessors have been executed
                            predecessors = self._get_predecessors(next_node)
                            if all(pred.id in executed_nodes for pred in predecessors):
                                execution_queue.append(next_node)
                    
                    elif isinstance(node, (DecisionFunc, DecisionLLM)):
                        # Evaluate the decision
                        branch = self._evaluate_decision(node, self.state, verbose)
                        executed_nodes.add(node.id)
                        
                        # Add the appropriate branch to the execution queue
                        if branch:
                            if isinstance(branch, TaskNode):
                                # Make sure the task in the branch has context from previous outputs
                                if not branch.task.context and self.state.task_outputs:
                                    latest_output = self.state.get_latest_output()
                                    if latest_output:
                                        branch.task.context = [latest_output]
                                        if verbose:
                                            console.print(f"[dim]Setting context from previous output for branch task: {escape_rich_markup(branch.task.description)}[/dim]")
                                
                                execution_queue.insert(0, branch)
                            elif isinstance(branch, TaskChain):
                                # Add all nodes from the branch to the queue in reverse order
                                branch_nodes = list(reversed(branch.nodes))
                                
                                # Set context for the first task in the branch if it doesn't have context
                                if branch_nodes and isinstance(branch_nodes[-1], TaskNode):
                                    first_task = branch_nodes[-1]
                                    if not first_task.task.context and self.state.task_outputs:
                                        latest_output = self.state.get_latest_output()
                                        if latest_output:
                                            first_task.task.context = [latest_output]
                                            if verbose:
                                                console.print(f"[dim]Setting context from previous output for first branch chain task: {escape_rich_markup(first_task.task.description)}[/dim]")
                                
                                for branch_node in branch_nodes:
                                    execution_queue.insert(0, branch_node)
                            elif isinstance(branch, (DecisionFunc, DecisionLLM)):
                                # Handle the case where a decision returns another decision node
                                # Insert at the front of the queue to be processed next
                                if verbose:
                                    console.print(f"[dim]Decision returned another decision node: {escape_rich_markup(branch.description)}[/dim]")
                                execution_queue.insert(0, branch)
                    
                    # Increment completed nodes and update progress
                    completed_nodes += 1
                    progress.update(overall_task, completed=completed_nodes)
                
                # Make sure the progress bar reaches 100% at the end
                progress.update(overall_task, completed=all_nodes)
        else:
            # Process nodes without progress bar
            while execution_queue:
                node = execution_queue.pop(0)
                
                # Process the node
                if isinstance(node, TaskNode):
                    # If this task doesn't have context but there are previous outputs, add them as context
                    if not node.task.context and self.state.task_outputs:
                        # Get the latest output to use as context
                        latest_output = self.state.get_latest_output()
                        if latest_output:
                            # Set the context for this task
                            node.task.context = [latest_output]
                    
                    output = self._execute_task(node, self.state, verbose)
                    self.state.update(node.id, output)
                    executed_nodes.add(node.id)
                    
                    # Add successor nodes to the queue
                    for next_node in self._get_next_nodes(node):
                        # Only add if all predecessors have been executed
                        predecessors = self._get_predecessors(next_node)
                        if all(pred.id in executed_nodes for pred in predecessors):
                            execution_queue.append(next_node)
                
                elif isinstance(node, (DecisionFunc, DecisionLLM)):
                    # Evaluate the decision
                    branch = self._evaluate_decision(node, self.state, verbose)
                    executed_nodes.add(node.id)
                    
                    # Add the appropriate branch to the execution queue
                    if branch:
                        if isinstance(branch, TaskNode):
                            # Make sure the task in the branch has context from previous outputs
                            if not branch.task.context and self.state.task_outputs:
                                latest_output = self.state.get_latest_output()
                                if latest_output:
                                    branch.task.context = [latest_output]
                                    
                            execution_queue.insert(0, branch)
                        elif isinstance(branch, TaskChain):
                            # Add all nodes from the branch to the queue in reverse order
                            branch_nodes = list(reversed(branch.nodes))
                            
                            # Set context for the first task in the branch if it doesn't have context
                            if branch_nodes and isinstance(branch_nodes[-1], TaskNode):
                                first_task = branch_nodes[-1]
                                if not first_task.task.context and self.state.task_outputs:
                                    latest_output = self.state.get_latest_output()
                                    if latest_output:
                                        first_task.task.context = [latest_output]
                            
                            for branch_node in branch_nodes:
                                execution_queue.insert(0, branch_node)
                        elif isinstance(branch, (DecisionFunc, DecisionLLM)):
                            # Handle the case where a decision returns another decision node
                            # Insert at the front of the queue to be processed next
                            if verbose:
                                console.print(f"[dim]Decision returned another decision node: {escape_rich_markup(branch.description)}[/dim]")
                            execution_queue.insert(0, branch)
        
        if verbose:
            console.print("[bold green]Graph Execution Completed[/bold green]")
            spacing()
        
        return self.state
    
    def _count_all_possible_nodes(self) -> int:
        """
        Counts all possible nodes in the graph, including all branches.
        
        Returns:
            The total number of nodes in the graph
        """
        # Start with the nodes directly in the graph
        counted = set()
        to_count = list(self.nodes)
        
        while to_count:
            node = to_count.pop(0)
            if node.id in counted:
                continue
                
            counted.add(node.id)
            
            # If it's a decision node, add its branches to be counted
            if isinstance(node, (DecisionFunc, DecisionLLM)):
                if node.true_branch:
                    if isinstance(node.true_branch, TaskNode):
                        if node.true_branch.id not in counted:
                            to_count.append(node.true_branch)
                    elif isinstance(node.true_branch, TaskChain):
                        for branch_node in node.true_branch.nodes:
                            if branch_node.id not in counted:
                                to_count.append(branch_node)
                
                if node.false_branch:
                    if isinstance(node.false_branch, TaskNode):
                        if node.false_branch.id not in counted:
                            to_count.append(node.false_branch)
                    elif isinstance(node.false_branch, TaskChain):
                        for branch_node in node.false_branch.nodes:
                            if branch_node.id not in counted:
                                to_count.append(branch_node)
        
        # Return the count, minimum of 1 to avoid division by zero
        return max(len(counted), 1)
    
    def run(self, verbose: bool = True, show_progress: bool = None) -> State:
        """
        Executes the graph, running all tasks in the appropriate order.
        
        Args:
            verbose: Whether to print detailed information
            show_progress: Whether to display a progress bar during execution. If None, uses the graph's show_progress attribute.
            
        Returns:
            The final state object with all task outputs
        """
        # Use class attribute if show_progress is not explicitly specified
        if show_progress is None:
            show_progress = self.show_progress
            
        if verbose:
            console.print("[bold blue]Starting Graph Execution[/bold blue]")
            spacing()
        
        # Reset state
        self.state = State()
        
        # With decision support, we always use the sequential implementation for now
        return self._run_sequential(verbose, show_progress)
    
    def get_output(self) -> Any:
        """
        Gets the output of the last task executed in the graph.
        
        Returns:
            The output of the last task
        """
        return self.state.get_latest_output()
    
    def get_task_output(self, description: str) -> Any:
        """
        Gets the output of a task by its description.
        
        Args:
            description: The description of the task
            
        Returns:
            The output of the specified task, or None if not found
        """
        for node in self.nodes:
            if isinstance(node, TaskNode) and node.task.description == description:
                output = self.state.get_task_output(node.id)
                if output is not None:
                    return output
        
        # If not found in direct nodes, check for nodes in decision branches
        for node in self.nodes:
            if isinstance(node, (DecisionFunc, DecisionLLM)):
                # Check true branch
                if node.true_branch:
                    if isinstance(node.true_branch, TaskNode) and node.true_branch.task.description == description:
                        return self.state.get_task_output(node.true_branch.id)
                    elif isinstance(node.true_branch, TaskChain):
                        for branch_node in node.true_branch.nodes:
                            if isinstance(branch_node, TaskNode) and branch_node.task.description == description:
                                return self.state.get_task_output(branch_node.id)
                
                # Check false branch
                if node.false_branch:
                    if isinstance(node.false_branch, TaskNode) and node.false_branch.task.description == description:
                        return self.state.get_task_output(node.false_branch.id)
                    elif isinstance(node.false_branch, TaskChain):
                        for branch_node in node.false_branch.nodes:
                            if isinstance(branch_node, TaskNode) and branch_node.task.description == description:
                                return self.state.get_task_output(branch_node.id)
        
        return None


# Helper functions to work with the existing Task class
def task(description: str, **kwargs) -> Task:
    """
    Creates a new Task with the given description and parameters.
    
    Args:
        description: The description of the task
        **kwargs: Additional parameters for the Task
        
    Returns:
        A new Task instance
    """
    # Ensure agent is explicitly set to None if not provided
    if 'agent' not in kwargs:
        kwargs['agent'] = None
    return Task(description=description, **kwargs)

def node(task_instance: Task) -> TaskNode:
    """
    Creates a new TaskNode wrapping the given Task.
    
    Args:
        task_instance: The Task to wrap
        
    Returns:
        A new TaskNode instance
    """
    return TaskNode(task=task_instance)

def create_graph(default_agent: Optional[Any] = None,
                 parallel_execution: bool = False,
                 show_progress: bool = True) -> Graph:
    """
    Creates a new graph with the specified configuration.
    
    Args:
        default_agent: Default agent to use for tasks (AgentConfiguration or Direct)
        parallel_execution: Whether to execute independent tasks in parallel
        show_progress: Whether to display a progress bar during execution
        
    Returns:
        A configured Graph instance
    """
    return Graph(
        default_agent=default_agent,
        parallel_execution=parallel_execution,
        show_progress=show_progress
    )


# Enable Task objects to use the >> operator directly
def _task_rshift(self, other):
    """
    Implements the >> operator for Task objects to connect them in a chain.
    
    Args:
        other: The next task in the chain
        
    Returns:
        A TaskChain object containing both tasks as nodes
    """
    chain = TaskChain()
    chain.add(TaskNode(task=self))
    
    if isinstance(other, Task):
        chain.add(TaskNode(task=other))
    elif isinstance(other, TaskNode):
        chain.add(other)
    elif isinstance(other, TaskChain):
        chain = chain.add(other)
    elif isinstance(other, (DecisionFunc, DecisionLLM)):
        chain.add(other)
    
    return chain

# Apply the patch to the Task class
Task.__rshift__ = _task_rshift
